//
// Blobs screensaver
// -------------------
// (c) 1996 Lucian Wischik
// Portions (c) 1991-1995 Borland International
//
// TBlobsData is the name of the class that holds configuration
// data as well as the current active state of the saver.
// Note that a new, separate instances of TBlobsData are used
// for the configuration dialog and for the screensaver.
// In response to a PSN_APPLY, the dialog procedure reads its
// current state, writes it to the registry, and routes a
// SCRM_CHANGE message in case it was being run as an active config.
// Active Config is when, while the screensaver is running, you
// press 'control' or the right mouse button to bring up the settings
// from within the screensaver.
// In response to a SCRM_CHANGE message, the main screensaver window
// deletes its old copy of TBlobsData and creates a new one that
// obtains its configuration from the newly changed registry.



#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <stdlib.h>
#pragma hdrstop

#include "../../lib/scrplus.h"
#include "blobs.rh"


// Note: I've put the following #defines just so this example compiles
// even if you don't have the latest headers
#if (WINVER < 0x0400)
#define WM_NOTIFY 0x004E
typedef struct tagNMHDR
{ HWND  hwndFrom;
  UINT  idFrom;
  UINT  code;         // NM_ code
} NMHDR;
typedef NMHDR FAR * LPNMHDR;
#define PSN_FIRST (0U-200U)
#define PSN_APPLY (PSN_FIRST-2)
#endif


const int NumThings = 900;
const int MaxBuilders = 8;

const int SCRCHANGE_BLOBS=2;


//
// An object on the screen
//
class TScreenThing {
  public:
    TScreenThing(SIZE& scrnSize);
    virtual void Draw(HDC hdc) = 0;
  protected:
    const POINT& Position() {return Pos;}
    const COLORREF& Color() {return C;}
  private:
    POINT  Pos;
    COLORREF  C;
};

typedef TScreenThing* (*TScreenThingBuilder)(SIZE&);

//
//
//
TScreenThing::TScreenThing(SIZE& scrnSize)
{
  Pos.x = rand() % (scrnSize.cx + 1);
  Pos.y = rand() % (scrnSize.cy + 1);
  C = RGB(rand() % 0x100, rand() % 0x100, rand() % 0x100);
}


//
// A screen thing that is a small dot
//
class TDot : public TScreenThing {
  public:
    TDot(SIZE& scrnSize) : TScreenThing(scrnSize) {}
    virtual void Draw(HDC hdc);
    static TScreenThing* Build(SIZE& scrnSize) {return new TDot(scrnSize);}
};

//
//
//
void
TDot::Draw(HDC hdc)
{
  SetPixel(hdc,Position().x,Position().y,Color());
}

//
// A screen thing that is a line segment
//
class TLine : public TScreenThing  {
  public:
    TLine(SIZE& scrnSize) : TScreenThing(scrnSize) {
      End.x = rand() % (scrnSize.cx + 1);
      End.y = rand() % (scrnSize.cy + 1);
    }
    virtual void Draw(HDC dc);
	 static TScreenThing* Build(SIZE& scrnSize) {return new TLine(scrnSize);}
  private:
    POINT  End;
};

//
//
//
void
TLine::Draw(HDC hdc)
{
  HGDIOBJ holdpen=SelectObject(hdc,CreatePen(PS_SOLID,0,Color()));
  MoveToEx(hdc,Position().x,Position().y,NULL);
  LineTo(hdc,End.x,End.y);
  holdpen=SelectObject(hdc,holdpen);
  DeleteObject(holdpen);
}

//
// A larger circle screen thing
//
class TBlob : public TScreenThing  {
  public:
    TBlob(SIZE& scrnSize) : TScreenThing(scrnSize) {
      Rad.cx = rand() % (scrnSize.cx/20);
      Rad.cy = rand() % (scrnSize.cy/20);
    }

    virtual void Draw(HDC hdc);
    static TScreenThing* Build(SIZE& scrnSize) {return new TBlob(scrnSize);}

  private:
    SIZE  Rad;
};

//
//
//
void
TBlob::Draw(HDC hdc)
{
  LOGBRUSH lb; lb.lbStyle=BS_SOLID; lb.lbColor=Color(); lb.lbHatch=0;
  HGDIOBJ holdbrush=SelectObject(hdc,CreateBrushIndirect(&lb));
  Ellipse(hdc,Position().x-Rad.cx,Position().y-Rad.cy,
             Position().x+Rad.cx, Position().y+Rad.cy);
  holdbrush=SelectObject(hdc,holdbrush);
  DeleteObject(holdbrush);
}

//----------------------------------------------------------------------------

//
//
//
class TBlobsData {
  public:
    TBlobsData();
    ~TBlobsData();
    void Seed(HWND);
    void AnimateScreen(HDC);
    void GetControls(HWND);
    void SetControls(HWND);
    void ReadRegistry();
	 void WriteRegistry();
    void NotifyChange(HWND);
    UINT uTimer;
  private:
    TScreenThing*  Thing[NumThings];
    TScreenThingBuilder Builder[MaxBuilders];
    SIZE           ScreenSize;
    int            DrawIndex, EraseIndex;
    int            NumBuilders;
	 BOOL           DoDots;
    BOOL           DoLines;
    BOOL           DoBlobs;
    int            speed;
};

//
//
//
TBlobsData::TBlobsData()
{ for (int i = 0; i < NumThings; i++)
    Thing[i] = 0;
  DoDots=TRUE; DoLines=TRUE; DoBlobs=TRUE;
}

void TBlobsData::Seed(HWND hwnd)
{
  DrawIndex  = 0;
  EraseIndex = -2*(NumThings/3);
  RECT rc; GetWindowRect(hwnd,&rc);
  ScreenSize.cx = rc.right-rc.left;
  ScreenSize.cy = rc.bottom-rc.top;

  speed=2;

  NumBuilders = 0;
  if (DoDots)
    Builder[NumBuilders++] = TDot::Build;
  if (DoLines)
    Builder[NumBuilders++] = TLine::Build;
  if (DoBlobs)
    Builder[NumBuilders++] = TBlob::Build;
}

//
//
//
TBlobsData::~TBlobsData()
{
  for (int i = 0; i < NumThings; i++)
  { if (Thing[i]!=NULL) delete Thing[i]; Thing[i]=NULL; }
}


//
//
//
void
TBlobsData::AnimateScreen(HDC hdc)
{
  if (!DoDots && !DoLines && !DoBlobs)
    return;

  int oldrop2=SetROP2(hdc,R2_XORPEN);

  Thing[DrawIndex] = Builder[rand() % (NumBuilders)](ScreenSize);
  Thing[DrawIndex]->Draw(hdc);

  DrawIndex++;
  DrawIndex %= NumThings;

  if (EraseIndex >= 0) {
    Thing[EraseIndex]->Draw(hdc);
	 delete Thing[EraseIndex];
    Thing[EraseIndex] = 0;
    EraseIndex++;
    EraseIndex %= NumThings;
  } else
    EraseIndex++;

  SetROP2(hdc,oldrop2);

}






LRESULT CALLBACK MyScreenSaverProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TBlobsData *bd=(TBlobsData *)GetWindowLong(hwnd,GWL_USERDATA);
  if (bd==NULL && msg!=WM_CREATE) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  switch (msg) {
    case (WM_CREATE):
		{ TBlobsData *bd=new TBlobsData();
        bd->ReadRegistry();
        bd->Seed(hwnd);
        SetWindowLong(hwnd,GWL_USERDATA,(LONG)bd);
        bd->uTimer=SetTimer(hwnd,1,20,NULL);
      } break;
    case (WM_ERASEBKGND):
      { RECT rc; HDC hdc=GetDC(hwnd); GetClientRect(hwnd,&rc);
        FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH)); ReleaseDC(hwnd,hdc);
      } break;
    case (WM_TIMER):
      { HDC hdc=GetDC(hwnd);
		  bd->AnimateScreen(hdc);
        ReleaseDC(hwnd,hdc);
      } break;
    case (WM_DESTROY):
      { if (bd->uTimer) KillTimer(hwnd,bd->uTimer);
        delete bd;
        SetWindowLong(hwnd,GWL_USERDATA,0);
      } break;
    case (SCRM_CHANGE):
		{ delete bd;
        bd=new TBlobsData();
        bd->ReadRegistry();
        bd->Seed(hwnd);
        SetWindowLong(hwnd,GWL_USERDATA,(LONG)bd);
        InvalidateRect(hwnd,NULL,TRUE);
      }
  }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}



void TBlobsData::SetControls(HWND hwnd)
{ CheckDlgButton(hwnd,ID_DOLINES,DoLines);
  CheckDlgButton(hwnd,ID_DODOTS,DoDots);
  CheckDlgButton(hwnd,ID_DOBLOBS,DoBlobs);
}

void TBlobsData::GetControls(HWND hwnd)
{ DoLines=IsDlgButtonChecked(hwnd,ID_DOLINES);
  DoDots=IsDlgButtonChecked(hwnd,ID_DODOTS);
  DoBlobs=IsDlgButtonChecked(hwnd,ID_DOBLOBS);
}

void TBlobsData::ReadRegistry()
{ LONG res; HKEY skey; DWORD val,valsize,valtype;
  res=RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Blobs",0,KEY_ALL_ACCESS,&skey);
  if (res!=ERROR_SUCCESS) return;
  // 1. dolines
  valsize=sizeof(val); res=RegQueryValueEx(skey,"DoLines",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) DoLines=val;
  // 2. dodots
  valsize=sizeof(val); res=RegQueryValueEx(skey,"DoDots",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) DoDots=(BOOL) val;
  // 3. doblobs
  valsize=sizeof(val); res=RegQueryValueEx(skey,"DoBlobs",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) DoBlobs=(BOOL) val;
  RegCloseKey(skey);
}

void TBlobsData::WriteRegistry()
{ LONG res; HKEY skey; DWORD val; DWORD disp;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Blobs",0,NULL,
                     REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  // 1. Dolines
  val=DoLines;
  RegSetValueEx(skey,"DoLines",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 2. Dodots
  val=DoDots;
  RegSetValueEx(skey,"DoDots",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 3. Doblobs
  val=DoBlobs;
  RegSetValueEx(skey,"DoBlobs",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
}

void TBlobsData::NotifyChange(HWND hwnd)
{ UINT code=SCRCHANGE_BLOBS | SCRCHANGE_FROMAPPLY   | SCRCHANGE_TOFULL;
  TScrChangeRouter cr;
  cr.hsrcwnd=hwnd; cr.lParam=(LPARAM)this; cr.code=code;
  HWND hdlg=GetParent(hwnd);
  SendMessage(hdlg,SCRM_CHANGE,0,(LPARAM)&cr);
}


BOOL WINAPI ConfigProc(HWND hwnd, UINT msg,WPARAM,LPARAM lParam)
{ TBlobsData *bd=(TBlobsData *)GetWindowLong(hwnd,DWL_USER);
  if (bd==NULL && msg!=WM_INITDIALOG) return FALSE;
  switch (msg) {
    case WM_INITDIALOG:
    { TBlobsData *bd=new TBlobsData();
      SetWindowLong(hwnd,DWL_USER,(LONG)bd);
      bd->ReadRegistry();
		bd->SetControls(hwnd);
      return TRUE;
    }
    case WM_DESTROY:
    { delete bd;
      SetWindowLong(hwnd,DWL_USER,0);
      return TRUE;
	 }
	 case (WM_NOTIFY):
	 { LPNMHDR nmh=(LPNMHDR) lParam;
		UINT code=nmh->code;
		switch (code) {
		  case (PSN_APPLY):
		  { bd->GetControls(hwnd);
			 bd->WriteRegistry();
			 bd->NotifyChange(GetParent(hwnd));
			 return TRUE;
		  }
		  default: return FALSE;
		}
	 }
  }
  return FALSE;
}







int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE,LPSTR,int)
{ // The following few lines are to set up global preferences
  TScrPrefs *p=ScrCreatePrefs();
  p->AllowActiveConfig=TRUE;
  p->AllowMute=FALSE;
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
  ScrSetPrefs(p);
  // And this next line runs the screen saver!
  int res=ScrExecute(hInstance,MyScreenSaverProc,ConfigProc);
  return res;
}


