//
// Mouse screensaver
// -------------------
// (c) 1998 Lucian Wischik
//
// This saver is based largely on the BlobsVC saver.
// But a few bits have been added to it to make it interactive
// with the mouse: it shows a cursor, and you can click in places
// to make it display only a certain type of blob, and you can
// click in the bottom right to pop up a dialog, and the only
// way to quit it is by clicking the right mouse button, and
// just before it terminates it pops up a 'goodbye' dialog. Discussion
// of how to achieve these features is in the MyScreenSaverProc function. 
//
// Behaviour if the user has not enabled passwords:
//      the saver runs as normal and allows interraction, and then
//      after the user right-clicks and dismisses the goodbye dialog.
//
// Behaviour if the user has enabled passwords:
// '95. The user right-clicks. This leads to a WM_CLOSE. The goodbye
//      dialog comes up. Then the DefScreenSaverProc gets called to
//      handle WM_CLOSE. This will do the password dialog box and
//      either destroy the window (and terminate the saver), or it will
//      just swallow it, depending on whether the user got the password
//      right.
// NT.  The user right-clicks. This leads to a WM_CLOSE. The goodbye
//      dialog comes up. Then the DefScreenSaverProc gets called to
//      handle WM_CLOSE, but because this is NT it obviously can't check
//      the password itself. The screensaver terminates and a password
//      dialog appears over the background wallpaper. If the user gets
//      the password right, the regular desktop reappears. If the user
//      gets it wrong, or doesn't type anything, then after a minute or
//      two the saver will start again from scratch.
//
// If you want exactly the same behaviour under NT as under '95, or if you
// want to put in your own password-validation stuff, then you must disable
// the official passwords (using the registry; see the html help for details)
// and put your own password dialog in. WM_CLOSE would be a good place.


#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <stdlib.h>
#pragma hdrstop

#include "../../lib/scrplus.h"
#include "mouse.rh"


// Note: I've put the following #defines just so this example compiles
// even if you don't have the latest headers
#if (WINVER < 0x0400)
#define WM_NOTIFY 0x004E
typedef struct tagNMHDR
{ HWND  hwndFrom;
  UINT  idFrom;
  UINT  code;         // NM_ code
} NMHDR;
typedef NMHDR FAR * LPNMHDR;
#define PSN_FIRST (0U-200U)
#define PSN_APPLY (PSN_FIRST-2)
#endif


const int NumThings = 900;
const int MaxBuilders = 8;

const int SCRCHANGE_BLOBS=2;


//
// An object on the screen
//
class TScreenThing {
  public:
    TScreenThing(SIZE& scrnSize);
    virtual void Draw(HDC hdc) = 0;
  protected:
    const POINT& Position() {return Pos;}
    const COLORREF& Color() {return C;}
  private:
    POINT  Pos;
    COLORREF  C;
};

typedef TScreenThing* (*TScreenThingBuilder)(SIZE&);

//
//
//
TScreenThing::TScreenThing(SIZE& scrnSize)
{
  Pos.x = rand() % (scrnSize.cx + 1);
  Pos.y = rand() % (scrnSize.cy + 1);
  C = RGB(rand() % 0x100, rand() % 0x100, rand() % 0x100);
}


//
// A screen thing that is a small dot
//
class TDot : public TScreenThing {
  public:
    TDot(SIZE& scrnSize) : TScreenThing(scrnSize) {}
    virtual void Draw(HDC hdc);
    static TScreenThing* Build(SIZE& scrnSize) {return new TDot(scrnSize);}
};

//
//
//
void
TDot::Draw(HDC hdc)
{
  SetPixel(hdc,Position().x,Position().y,Color());
}

//
// A screen thing that is a line segment
//
class TLine : public TScreenThing  {
  public:
    TLine(SIZE& scrnSize) : TScreenThing(scrnSize) {
      End.x = rand() % (scrnSize.cx + 1);
      End.y = rand() % (scrnSize.cy + 1);
    }
    virtual void Draw(HDC dc);
	 static TScreenThing* Build(SIZE& scrnSize) {return new TLine(scrnSize);}
  private:
    POINT  End;
};

//
//
//
void
TLine::Draw(HDC hdc)
{
  HGDIOBJ holdpen=SelectObject(hdc,CreatePen(PS_SOLID,0,Color()));
  MoveToEx(hdc,Position().x,Position().y,NULL);
  LineTo(hdc,End.x,End.y);
  holdpen=SelectObject(hdc,holdpen);
  DeleteObject(holdpen);
}

//
// A larger circle screen thing
//
class TBlob : public TScreenThing  {
  public:
    TBlob(SIZE& scrnSize) : TScreenThing(scrnSize) {
      Rad.cx = rand() % (scrnSize.cx/20);
      Rad.cy = rand() % (scrnSize.cy/20);
    }

    virtual void Draw(HDC hdc);
    static TScreenThing* Build(SIZE& scrnSize) {return new TBlob(scrnSize);}

  private:
    SIZE  Rad;
};

//
//
//
void
TBlob::Draw(HDC hdc)
{
  LOGBRUSH lb; lb.lbStyle=BS_SOLID; lb.lbColor=Color(); lb.lbHatch=0;
  HGDIOBJ holdbrush=SelectObject(hdc,CreateBrushIndirect(&lb));
  Ellipse(hdc,Position().x-Rad.cx,Position().y-Rad.cy,
             Position().x+Rad.cx, Position().y+Rad.cy);
  holdbrush=SelectObject(hdc,holdbrush);
  DeleteObject(holdbrush);
}

//----------------------------------------------------------------------------

//
//
//
class TBlobsData {
  public:
    TBlobsData();
    ~TBlobsData();
    void Seed(HWND);
    void AnimateScreen(HDC);
    void GetControls(HWND);
    void SetControls(HWND);
    void ReadRegistry();
	 void WriteRegistry();
    void NotifyChange(HWND);
    UINT uTimer;
	 BOOL           DoDots;
    BOOL           DoLines;
    BOOL           DoBlobs;
  private:
    TScreenThing*  Thing[NumThings];
    TScreenThingBuilder Builder[MaxBuilders];
    SIZE           ScreenSize;
    int            DrawIndex, EraseIndex;
    int            NumBuilders;
    int            speed;
};

//
//
//
TBlobsData::TBlobsData()
{ for (int i = 0; i < NumThings; i++)
    Thing[i] = 0;
  DoDots=TRUE; DoLines=TRUE; DoBlobs=TRUE;
}

void TBlobsData::Seed(HWND hwnd)
{
  DrawIndex  = 0;
  EraseIndex = -2*(NumThings/3);
  RECT rc; GetWindowRect(hwnd,&rc);
  ScreenSize.cx = rc.right-rc.left;
  ScreenSize.cy = rc.bottom-rc.top;

  speed=2;

  NumBuilders = 0;
  if (DoDots)
    Builder[NumBuilders++] = TDot::Build;
  if (DoLines)
    Builder[NumBuilders++] = TLine::Build;
  if (DoBlobs)
    Builder[NumBuilders++] = TBlob::Build;
}

//
//
//
TBlobsData::~TBlobsData()
{
  for (int i = 0; i < NumThings; i++)
  { if (Thing[i]!=NULL) delete Thing[i]; Thing[i]=NULL; }
}


//
//
//
void
TBlobsData::AnimateScreen(HDC hdc)
{
  if (!DoDots && !DoLines && !DoBlobs)
    return;

  int oldrop2=SetROP2(hdc,R2_XORPEN);

  Thing[DrawIndex] = Builder[rand() % (NumBuilders)](ScreenSize);
  Thing[DrawIndex]->Draw(hdc);

  DrawIndex++;
  DrawIndex %= NumThings;

  if (EraseIndex >= 0) {
    Thing[EraseIndex]->Draw(hdc);
	 delete Thing[EraseIndex];
    Thing[EraseIndex] = 0;
    EraseIndex++;
    EraseIndex %= NumThings;
  } else
    EraseIndex++;

  SetROP2(hdc,oldrop2);

}





BOOL reallyclose=FALSE;

LRESULT CALLBACK MyScreenSaverProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TBlobsData *bd=(TBlobsData *)GetWindowLong(hwnd,GWL_USERDATA);
  if (bd==NULL && msg!=WM_CREATE) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  switch (msg) {
    case (WM_CREATE):
		{ TBlobsData *bd=new TBlobsData();
        bd->ReadRegistry();
        bd->Seed(hwnd);
        SetWindowLong(hwnd,GWL_USERDATA,(LONG)bd);
        bd->uTimer=SetTimer(hwnd,1,20,NULL);
      } break;
    case (WM_ERASEBKGND):
      { RECT rc; HDC hdc=GetDC(hwnd); GetClientRect(hwnd,&rc);
        FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH));
		char *c="Click left button in top-left, top-right or bottom-left to change blob type";
		TextOut(hdc,16,48,c,strlen(c));
		c="Click left button in bottom-right to pop up a dialog";
		TextOut(hdc,16,64,c,strlen(c));
		c="Click right button to get out of the saver";
		TextOut(hdc,16,96,c,strlen(c));
		ReleaseDC(hwnd,hdc);
      } break;
    case (WM_TIMER):
      { HDC hdc=GetDC(hwnd);
		  bd->AnimateScreen(hdc);
        ReleaseDC(hwnd,hdc);
      } break;
    case (WM_DESTROY):
      {  
         if (bd->uTimer) KillTimer(hwnd,bd->uTimer);
        delete bd;
        SetWindowLong(hwnd,GWL_USERDATA,0);
      } break;
	//
	// THE FOLLOWING THINGS WERE ADDED TO GET THE INTERACTIVE MOUSE SUPPORT STUFF
	//
	case (WM_MOUSEMOVE):
		{ MessageBeep(0); return 0;
		  // just to show off that we can respond to mouse moves
		} break;
	case (WM_SETCURSOR):
		{ SetCursor(LoadCursor(NULL,IDC_SIZEALL)); return 0;
		  // because otherwise there would be no cursor
		} break;
	case (WM_LBUTTONDOWN):
		{ // in response to a left-button click:
		  // if the user clicked top-left, top-right or bottom-left, we change the behaviour
		  // of the saver so that it only draws a particular type of blob.
		  // if the user clicked bottom-right, then a dialog pops up.
		  // Note our use of SendMessage(..,SCRM_DIALOG,..) to indicate to ScrPlus that
		  // the dialog belonged to us, and that its appearance and subsequent dismissal
		  // should not be taken as grounds for terminating the saver.
		  int x=LOWORD(lParam), y=HIWORD(lParam);
		  RECT rc; GetWindowRect(hwnd,&rc); int cx=rc.right-rc.left, cy=rc.bottom-rc.top;
		  int quadrant=((x*2<cx)?0:2)+((y*2<cy)?0:1);
		  bd->DoLines=FALSE; bd->DoBlobs=FALSE; bd->DoDots=FALSE;
		  switch (quadrant)
		  { case 0: bd->DoLines=TRUE; break;
		  case 1: bd->DoBlobs=TRUE; break;
		  case 2: bd->DoDots=TRUE; break;
		  case 3: bd->DoLines=TRUE; bd->DoBlobs=TRUE; bd->DoDots=TRUE;
			  SendMessage(hwnd,SCRM_DIALOG,TRUE,0);
			  MessageBox(hwnd,"message box","box",MB_OK);
			  SendMessage(hwnd,SCRM_DIALOG,FALSE,0);
			  break;
		  }
		  bd->Seed(hwnd);
		  return 0;
		}
	case (WM_RBUTTONDOWN):
		{ // If the user clicks right-button, we terminate the saver. (See below for discussion of
		  // reallyclose).
		  reallyclose=TRUE;
		  SendMessage(hwnd,WM_CLOSE,0,0);
		} break;
	case (WM_CLOSE):
		{ // In response to a WM_CLOSE message:
		  // Note that WindowsNT itself will tend to send us WM_CLOSE messages when it thinks that
          // a mouse-move or click or something should be enough to terminate the saver. We (yuck!)
		  // used a global boolean flag 'reallyclose' which will be false normally so that we can
		  // detect these Windows-generated close messages. Only when the user clicked the right-button
		  // will the reallyclose flag end up being set, and only in this case will the saver terminate.
		  // (Of course, these concerns are not relevant when the saver is running in a preview window.
		  // The trick to tell whether we're running as a preview or a full screen I copied out of a bit
		  // of ScrPlus source code. The idea is that the full-screen saver runs without a parent,
		  // while any previews do have a parent.
		  // ONE OTHER NEAT THING. Just before the saver closes we pop up a message box that says
		  // 'goodbye'. You might wish to put an advertising dialog box in here, or a nag screen or
		  // credits or somethting.
		  BOOL isfull=FALSE;
          if (GetParent(hwnd)==NULL) isfull=TRUE;
		  if (isfull && !reallyclose) return FALSE;
		  if (isfull) MessageBox(hwnd,"Just a final goodbye...","Goodbye",MB_OK);
		} break;
	
	//
 }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}






void TBlobsData::SetControls(HWND hwnd)
{ CheckDlgButton(hwnd,ID_DOLINES,DoLines);
  CheckDlgButton(hwnd,ID_DODOTS,DoDots);
  CheckDlgButton(hwnd,ID_DOBLOBS,DoBlobs);
}

void TBlobsData::GetControls(HWND hwnd)
{ DoLines=IsDlgButtonChecked(hwnd,ID_DOLINES);
  DoDots=IsDlgButtonChecked(hwnd,ID_DODOTS);
  DoBlobs=IsDlgButtonChecked(hwnd,ID_DOBLOBS);
}

void TBlobsData::ReadRegistry()
{ LONG res; HKEY skey; DWORD val,valsize,valtype;
  res=RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Blobs",0,KEY_ALL_ACCESS,&skey);
  if (res!=ERROR_SUCCESS) return;
  // 1. dolines
  valsize=sizeof(val); res=RegQueryValueEx(skey,"DoLines",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) DoLines=val;
  // 2. dodots
  valsize=sizeof(val); res=RegQueryValueEx(skey,"DoDots",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) DoDots=(BOOL) val;
  // 3. doblobs
  valsize=sizeof(val); res=RegQueryValueEx(skey,"DoBlobs",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) DoBlobs=(BOOL) val;
  RegCloseKey(skey);
}

void TBlobsData::WriteRegistry()
{ LONG res; HKEY skey; DWORD val; DWORD disp;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Blobs",0,NULL,
                     REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  // 1. Dolines
  val=DoLines;
  RegSetValueEx(skey,"DoLines",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 2. Dodots
  val=DoDots;
  RegSetValueEx(skey,"DoDots",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 3. Doblobs
  val=DoBlobs;
  RegSetValueEx(skey,"DoBlobs",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
}



BOOL WINAPI ConfigProc(HWND hwnd, UINT msg,WPARAM,LPARAM lParam)
{ TBlobsData *bd=(TBlobsData *)GetWindowLong(hwnd,DWL_USER);
  if (bd==NULL && msg!=WM_INITDIALOG) return FALSE;
  switch (msg) {
    case WM_INITDIALOG:
    { TBlobsData *bd=new TBlobsData();
      SetWindowLong(hwnd,DWL_USER,(LONG)bd);
      bd->ReadRegistry();
		bd->SetControls(hwnd);
      return TRUE;
    }
    case WM_DESTROY:
    { delete bd;
      SetWindowLong(hwnd,DWL_USER,0);
      return TRUE;
	 }
  }
  return FALSE;
}







int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE,LPSTR,int)
{ // The following few lines are to set up global preferences
  //TScrPrefs *p=ScrCreatePrefs();
  //p->AllowMute=FALSE;
  //p->IsTrace=FALSE;
  //p->IsDebug=FALSE;
  //ScrSetPrefs(p);
  // And this next line runs the screen saver!
  int res=ScrExecute(hInstance,MyScreenSaverProc,ConfigProc);
  return res;
}


