//
// Sperm Screensaver
// -------------------
// Revisions:
// February 1991 - Drew Oldbrich. Created 'xsperm.c' as a demonstration of
//                 animation under X.
// May 1991      - Ali Ozer. Wrapped up animation functoin in a NextStep
//                 View subclass
// August 1991   - Sam Streeper. Very minor changes to make it work as a
//                 screen saver module.
// December 1992 - Edward Hutchins. Ported to Windows 3.1 as an SPX
//                 screensaver module.
// August 1995   - Lucian Wischik. Ported to Win95
//                 including extensive rewrites of the user interface.
//
//
// How it does it:
// First, the super timer messages. You have to put in a response to the
// SCRM_TIMER message, and have it return the expected period.
// The configuration dialog box in this example has the following behaviour:
//   1. when you make a change on a control, the preview updates immediately.
//   2. if it happens to be an active-config, then the main screensaver is
//      only updated when you click 'ok'
// Firstly, you have to program the user-interface behaviour:
//   1. On WM_INITDIALOG, create a data object. Then do
//      ReadRegistry(), Mark(), SetControls().
//   2. On WM_DESTROY, delete the data object
//   3. on WM_COMMAND, call GetControls. If anything has changed, then
//      call NotifyChanges(FALSE)
//   4. On PSN_APPLY, do Mark(), WriteRegistry() and then NotifyChanges(TRUE)
//   5. On PSN_RESET, do Reset(). If there are any changes, do
//      NotifyChanges(FALSE)
// All of the functions are methods of the data object. What they do is
// as follows:
// ReadRegistry: reads values in from the registry
// WriteRegistry: writes values to the registry
// Mark: For each variable X, there is a variable markX which remembers
//   the values from some previous time. Mark remembers them by copying
//   all X into MarkX.
// Reset: restores all variables X back to their markX values, and return
//   TRUE if there were any changes
// SetControls: Setup the dialog. This is called only once.
// GetControls: go through all controls reading their values. Returns
//   TRUE if there were any changes.
// NotifyChanges: route an appropriate SCRM_CHANGE to the appropriate place.

#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#pragma hdrstop
#include "process.h"

#include "../../lib/scrplus.h"
#include "sperm.rh"


// Note: I've put the following #defines just so this example compiles
// even if you don't have the latest headers
#if (WINVER < 0x0400)
#define WM_NOTIFY 0x004E
typedef struct tagNMHDR
{ HWND  hwndFrom;
  UINT  idFrom;
  UINT  code;         // NM_ code
} NMHDR;
typedef NMHDR FAR * LPNMHDR;
#define PSN_FIRST (0U-200U)
#define PSN_APPLY (PSN_FIRST-2)
#define PSN_RESET (PSN_FIRST-3)
#endif



typedef LONG VECTOR[2];
#define SPERM_MAX 2000

// fixed point scale factor
#define VEC_1 (256L)
#define NORM(x) ((INT)((x)/VEC_1))

#define VEC_COPY( y, x )        y[0] = x[0], y[1] = x[1]
#define VEC_ADD( z, x, y )      z[0] = x[0] + y[0], z[1] = x[1] + y[1]
#define VEC_SUB( z, x, y )      z[0] = x[0] - y[0], z[1] = x[1] - y[1]
#define VEC_MULT( x, a )        (x[0] = (x[0] * a) / VEC_1, x[1] = (x[1] * a) / VEC_1)
#define VEC_DIV( x, a )         (x[0] = (x[0] * VEC_1) / a, x[1] = (x[1] * VEC_1) / a)
#define VEC_ADDS( z, x, a, y )  (z[0] = x[0] + (((a)*y[0])/VEC_1), z[1] = x[1] + (((a)*y[1])/VEC_1))
#define VEC_NORM( x )           { LONG _nLen = IntDist( x[0], x[1] ); VEC_DIV( x, _nLen ); }

#define MINRAD      (VEC_1 / 10)
#define RADSTEP     2
#define MAXRAD      (MINRAD * 6 * RADSTEP)
#define INITRAD     (MINRAD * 3 * RADSTEP)

// RandInt(n) returns an integer 0..n-1
#define RandInt(n) (arand( n ))

// degrees scaled to integer math
#define DEGREE_SIZE 256
#define DEGREE_MASK 255
#define DEGREE_MAX 0x4000 // this is now hardcoded into the lookup table...

// faster than MulDiv!
#define MULDEG(x,y) (((x)*(y))/DEGREE_MAX)

// DEG - convert an integer into a degree lookup index
#define DEG(x) ((WORD)(x)&DEGREE_MASK)

const int numcounts=12;
const int defcountindex=5;
const BOOL deferase=TRUE;
char *countname[numcounts]=
{ "Dead","British male","Mouse","Hedgehog","Squirrel","Rabbit","Cat","Dog",
  "Donkey","Bull","Elephant","Stallion" };
int countnumber[numcounts]=
{ 2,10,20,30,50,80,120,180,270,400,600,800 };



int nCos[DEGREE_SIZE] =
{
 16384,  16379,  16364,  16339,  16305,  16260,  16206,  16142,
 16069,  15985,  15892,  15790,  15678,  15557,  15426,  15286,
 15136,  14978,  14810,  14634,  14449,  14255,  14053,  13842,
 13622,  13395,  13159,  12916,  12665,  12406,  12139,  11866,
 11585,  11297,  11002,  10701,  10393,  10079,  9759,  9434,
 9102,  8765,  8423,  8075,  7723,  7366,  7005,  6639,
 6269,  5896,  5519,  5139,  4756,  4369,  3980,  3589,
 3196,  2801,  2404,  2005,  1605,  1205,  803,  402,
 0, -403, -804, -1206, -1606, -2006, -2405, -2802,
-3197, -3590, -3981, -4370, -4757, -5140, -5520, -5897,
-6270, -6640, -7006, -7367, -7724, -8076, -8424, -8766,
-9103, -9435, -9760, -10080, -10394, -10702, -11003, -11298,
-11586, -11867, -12140, -12407, -12666, -12917, -13160, -13396,
-13623, -13843, -14054, -14256, -14450, -14635, -14811, -14979,
-15137, -15287, -15427, -15558, -15679, -15791, -15893, -15986,
-16070, -16143, -16207, -16261, -16306, -16340, -16365, -16380,
-16384, -16380, -16365, -16340, -16306, -16261, -16207, -16143,
-16070, -15986, -15893, -15791, -15679, -15558, -15427, -15287,
-15137, -14979, -14811, -14635, -14450, -14256, -14054, -13843,
-13623, -13396, -13160, -12917, -12666, -12407, -12140, -11867,
-11586, -11298, -11003, -10702, -10394, -10080, -9760, -9435,
-9103, -8766, -8424, -8076, -7724, -7367, -7006, -6640,
-6270, -5897, -5520, -5140, -4757, -4370, -3981, -3590,
-3197, -2802, -2405, -2006, -1606, -1206, -804, -403,
 0,  402,  803,  1205,  1605,  2005,  2404,  2801,
 3196,  3589,  3980,  4369,  4756,  5139,  5519,  5896,
 6269,  6639,  7005,  7366,  7723,  8075,  8423,  8765,
 9102,  9434,  9759,  10079,  10393,  10701,  11002,  11297,
 11585,  11866,  12139,  12406,  12665,  12916,  13159,  13395,
 13622,  13842,  14053,  14255,  14449,  14634,  14810,  14978,
 15136,  15286,  15426,  15557,  15678,  15790,  15892,  15985,
 16069,  16142,  16206,  16260,  16305,  16339,  16364,  16379
};
int nSin[DEGREE_SIZE] =
{
 0,  402,  803,  1205,  1605,  2005,  2404,  2801,
 3196,  3589,  3980,  4369,  4756,  5139,  5519,  5896,
 6269,  6639,  7005,  7366,  7723,  8075,  8423,  8765,
 9102,  9434,  9759,  10079,  10393,  10701,  11002,  11297,
 11585,  11866,  12139,  12406,  12665,  12916,  13159,  13395,
 13622,  13842,  14053,  14255,  14449,  14634,  14810,  14978,
 15136,  15286,  15426,  15557,  15678,  15790,  15892,  15985,
 16069,  16142,  16206,  16260,  16305,  16339,  16364,  16379,
 16383,  16379,  16364,  16339,  16305,  16260,  16206,  16142,
 16069,  15985,  15892,  15790,  15678,  15557,  15426,  15286,
 15136,  14978,  14810,  14634,  14449,  14255,  14053,  13842,
 13622,  13395,  13159,  12916,  12665,  12406,  12139,  11866,
 11585,  11297,  11002,  10701,  10393,  10079,  9759,  9434,
 9102,  8765,  8423,  8075,  7723,  7366,  7005,  6639,
 6269,  5896,  5519,  5139,  4756,  4369,  3980,  3589,
 3196,  2801,  2404,  2005,  1605,  1205,  803,  402,
 0, -403, -804, -1206, -1606, -2006, -2405, -2802,
-3197, -3590, -3981, -4370, -4757, -5140, -5520, -5897,
-6270, -6640, -7006, -7367, -7724, -8076, -8424, -8766,
-9103, -9435, -9760, -10080, -10394, -10702, -11003, -11298,
-11586, -11867, -12140, -12407, -12666, -12917, -13160, -13396,
-13623, -13843, -14054, -14256, -14450, -14635, -14811, -14979,
-15137, -15287, -15427, -15558, -15679, -15791, -15893, -15986,
-16070, -16143, -16207, -16261, -16306, -16340, -16365, -16380,
-16384, -16380, -16365, -16340, -16306, -16261, -16207, -16143,
-16070, -15986, -15893, -15791, -15679, -15558, -15427, -15287,
-15137, -14979, -14811, -14635, -14450, -14256, -14054, -13843,
-13623, -13396, -13160, -12917, -12666, -12407, -12140, -11867,
-11586, -11298, -11003, -10702, -10394, -10080, -9760, -9435,
-9103, -8766, -8424, -8076, -7724, -7367, -7006, -6640,
-6270, -5897, -5520, -5140, -4757, -4370, -3981, -3590,
-3197, -2802, -2405, -2006, -1606, -1206, -804, -403
};



class TSpermData { public:
  int     CountIndex, markCountIndex;
  BOOL    Erase, markErase;
  BOOL    FlatOut, markFlatOut;
  // Those two were registry things. The rest are for running it.
  int     nSpermCnt;
  UINT    idTimer;
  DWORD   dwSeed;
  VECTOR  x[SPERM_MAX], xPrev[SPERM_MAX];
  VECTOR  v[SPERM_MAX];
  VECTOR  lCenter;
  long    lSin[SPERM_MAX];
  long    lCos[SPERM_MAX];
  long    lVel[SPERM_MAX];
  long    lRad;
  long    lDir;
  int     nRndCnt1, nRndCnt2;
  int     cx,cy;
  BOOL ignorechanges;
  // thread stuff

  TSpermData();
  void Seed(HWND);
  void Step(HDC,BOOL);
  long arand(long);
  long IntDist(long,long);
  void DoEffect(int);
  void ReadRegistry();
  void WriteRegistry();
  void NotifyChanges(HWND hwnd,BOOL isapply);
  void Mark();
  BOOL Reset();
  void SetControls(HWND);
  BOOL GetControls(HWND);
};

TSpermData::TSpermData()
{ CountIndex=defcountindex; Erase=deferase; ignorechanges=FALSE;
  FlatOut=TRUE;
}

// arand - pseudorandom number from 0 to x-1
long TSpermData::arand( long x )
{ dwSeed = dwSeed * 0x343fd + 0x269ec3;
  return( (long)((dwSeed >> 8) % x) );
}

// IntDist - find an approximate distance
long TSpermData::IntDist(long dx,long dy )
{ long lDist;
  if (dx < 0) dx = -dx;  if (dy < 0) dy = -dy;
  lDist = dx + dy - (((dx > dy) ? dy : dx) >> 1);
  return( (lDist > 0) ? lDist : 1 );
}

// DoEffect - do a special effect
void TSpermData::DoEffect(int nVal )
{ int i;
  switch (nVal) {
    case 0:
	 { lDir *= -1;
		for (i = 0; i < nSpermCnt; i++)
		{ VECTOR y;
		  VEC_COPY( y, v[i] );
		  if (lDir < 0) v[i][0] = y[1], v[i][1] = -y[0];
		  else v[i][0] = -y[1], v[i][1] = y[0];
		}
    } break;
	case 1:
    { for (i = 0; i < nSpermCnt; i++) v[i][0] = -v[i][0];
    } break;
	case 2:
    { for (i = 0; i < nSpermCnt; i++) v[i][1] = -v[i][1];
    } break;
	case 3:
    { DoEffect( 1 );
		DoEffect( 2 );
    } break;
	case 4:
    { DoEffect( 0 );
		DoEffect( 3 );
    } // No break. Intentional?
	case 5:
    { lRad = lRad * RADSTEP;
		if (lRad > MAXRAD) lRad = MAXRAD;
    } break;
	case 6:
    { lRad = lRad / RADSTEP;
		if (lRad < MINRAD) lRad = MINRAD;
    } break;
  }
}

void TSpermData::Seed(HWND hwnd)
{	dwSeed += GetTickCount();
	lDir = 1;
	lRad = INITRAD;
   if (CountIndex<0) CountIndex=0;
   if (CountIndex>numcounts-1) CountIndex=numcounts-1;
   nSpermCnt=countnumber[CountIndex];
	if (nSpermCnt < 1) nSpermCnt = 1;
	else if (nSpermCnt >= SPERM_MAX) nSpermCnt = SPERM_MAX - 1;
	nRndCnt1 = 100;
	nRndCnt2 = 200;
   int i;
   RECT rc;
   GetWindowRect(hwnd,&rc);
   cx=rc.right-rc.left; cy=rc.bottom-rc.top;
	lCenter[0] = (cx / 2) * VEC_1;
	lCenter[1] = (cy / 2) * VEC_1;
	for (i = 0; i < nSpermCnt; i++)
	{  int nAngle = (int)RandInt( 10 ) + 5;
		xPrev[i][0] = x[i][0] = RandInt( cx * VEC_1 );
		xPrev[i][1] = x[i][1] = RandInt( cy * VEC_1 );
		v[i][0] = RandInt( 256 * VEC_1 ) - 128 * VEC_1;
		v[i][1] = RandInt( 256 * VEC_1 ) - 128 * VEC_1;
		VEC_NORM( v[i] );
		lSin[i] = nSin[ DEG(nAngle) ];
		lCos[i] = nCos[ DEG(nAngle) ];
		lVel[i] = RandInt( 4 * VEC_1 ) + 4 * VEC_1;
	}
}


void TSpermData::Step(HDC hdc, BOOL drawnew)
{ int i;

  HGDIOBJ hOldPen=SelectObject(hdc,GetStockObject(BLACK_PEN));
  for (i = 0; i < nSpermCnt; i++)
  { VECTOR      w, y;
	 VECTOR      p;
    VECTOR      xOld;
	 long        lTmp;

	 VEC_COPY( xOld, xPrev[i] );
	 VEC_COPY( xPrev[i], x[i] );

	 VEC_SUB( w, x[i], lCenter );
	 VEC_NORM( w );
	 VEC_COPY( y, w );
	 w[0] = MULDEG( y[0], lCos[i] ) - (lDir * MULDEG( y[1], lSin[i] )) / VEC_1;
	 w[1] = MULDEG( y[1], lCos[i] ) + (lDir * MULDEG( y[0], lSin[i] )) / VEC_1;
	 lTmp = (lRad * (160 * VEC_1 - lVel[i] * 20)) / VEC_1;
	 VEC_ADDS( p, lCenter, lTmp, w );
	 VEC_SUB( w, p, x[i] );
	 VEC_NORM( w );
	 VEC_ADD( v[i], v[i], w );
	 VEC_NORM( v[i] );
	 VEC_MULT( v[i], lVel[i] );
	 VEC_ADD( x[i], x[i], v[i] );

    SelectObject(hdc,GetStockObject(BLACK_PEN));
    MoveToEx( hdc, NORM(xOld[0]), NORM(xOld[1]),NULL );
    LineTo( hdc, NORM( xPrev[i][0] ), NORM( xPrev[i][1] ) );
    if (drawnew) SelectObject( hdc, GetStockObject( WHITE_PEN ) );
    LineTo( hdc, NORM( x[i][0] ), NORM( x[i][1] ) );
  }
  SelectObject(hdc,hOldPen);


  if (--nRndCnt1 < 0)
  { nRndCnt1 = (INT)RandInt( 700 );
  	 lCenter[0] = RandInt( cx * VEC_1 );
	 lCenter[1] = RandInt( cy * VEC_1 );
  }

  if (--nRndCnt2 < 0)
  { nRndCnt2 = (INT)RandInt( 600 );
	 DoEffect( nRndCnt2 % 7 );
  }

}

#define SCRCHANGE_SPERM 2



inline void SpermDebugString(char *c)
{ //return;
  char d[200];
  strcpy(d,c); strcat(d,"_______      ");
  OutputDebugString(d);
}



LRESULT CALLBACK MyScreenSaverProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TSpermData *sd=(TSpermData *)GetWindowLong(hwnd,GWL_USERDATA);
  if (sd==NULL && msg!=WM_CREATE) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  // Note: there's an outside possibility that this windowproc could be
  // called with a SCRM_GETLONG even AFTER! the WM_DESTROY had finished. The
  // above line means we don't have to worry about that.
  switch (msg) {
    case (WM_CREATE):
      { TSpermData *sd=new TSpermData();
        sd->ReadRegistry();
        sd->Seed(hwnd);
        SetWindowLong(hwnd,GWL_USERDATA,(LONG)sd);
      } break;
    case (WM_ERASEBKGND):
      { if (sd->Erase)
        { RECT rc; HDC hdc=(HDC)wParam; GetClientRect(hwnd,&rc);
          FillRect(hdc,&rc,(HBRUSH)GetStockObject(BLACK_BRUSH));
        }
      } break;
    case (SCRM_TIMER):
      { HDC hdc=GetDC(hwnd);
        sd->Step(hdc,TRUE);
		  ReleaseDC(hwnd,hdc);
        if (sd->FlatOut) return 1; else return 18;
      }
    case (SCRM_CHANGE):
      { if ((wParam & SCRCHANGE_GETCODE)==SCRCHANGE_SPERM)
        { TSpermData *od=(TSpermData *)lParam;
          if (od->Erase != sd->Erase)
          { sd->Erase=od->Erase;
            ShowWindow(hwnd,SW_HIDE);
            UpdateWindow(GetParent(hwnd));
            ShowWindow(hwnd,SW_SHOW);
          }
          if (od->CountIndex != sd->CountIndex)
          { sd->CountIndex=od->CountIndex;
            HDC hdc=GetDC(hwnd);
            sd->Step(hdc,FALSE); // to erase them all
            ReleaseDC(hwnd,hdc);
            sd->Seed(hwnd);
          }
          if (od->FlatOut != sd->FlatOut)
          { sd->FlatOut=od->FlatOut;
			 }
        }
      } break;
    case (WM_DESTROY):
      { SetWindowLong(hwnd,GWL_USERDATA,0);
        delete sd;
      } break;
  }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}




void TSpermData::SetControls(HWND hwnd)
{ ignorechanges=TRUE;
  for (int i=0; i<numcounts; i++)
  SendDlgItemMessage(hwnd,ID_SPERMCOUNT,CB_ADDSTRING,0,(LPARAM)countname[i]);
  SendDlgItemMessage(hwnd,ID_SPERMCOUNT,CB_SETCURSEL,CountIndex,0);
  CheckDlgButton(hwnd,ID_BLANK,Erase);
  CheckDlgButton(hwnd,ID_FLATOUT,FlatOut);
  ScrCreatePreview(GetDlgItem(hwnd,ID_MONITOR));
  ignorechanges=FALSE;
}

BOOL TSpermData::GetControls(HWND hwnd)
{ if (ignorechanges) return FALSE;
  BOOL ischanged=FALSE;
  LRESULT res=SendDlgItemMessage(hwnd,ID_SPERMCOUNT,CB_GETCURSEL,0,0);
  if (res!=CountIndex) ischanged=TRUE; CountIndex=res;
  UINT ru=IsDlgButtonChecked(hwnd,ID_BLANK);
  BOOL rb=TRUE; if (ru==0) rb=FALSE;
  if (rb!=Erase) ischanged=TRUE;
  Erase=rb;
  ru=IsDlgButtonChecked(hwnd,ID_FLATOUT);
  rb=TRUE; if (ru==0) rb=FALSE;
  if (rb!=FlatOut) ischanged=TRUE;
  FlatOut=rb;
  return ischanged;
}

BOOL TSpermData::Reset()
{ BOOL ischanged=FALSE;
  if (markCountIndex!=CountIndex) ischanged=TRUE; CountIndex=markCountIndex;
  if (markErase!=Erase) ischanged=TRUE; Erase=markErase;
  if (markFlatOut!=FlatOut) ischanged=TRUE; FlatOut=markFlatOut;
  return ischanged;
}

void TSpermData::Mark()
{ markCountIndex=CountIndex;
  markErase=Erase;
  markFlatOut=FlatOut;
}

void TSpermData::NotifyChanges(HWND hsrcwnd, BOOL isapply)
{ WPARAM code;
  if (isapply) code=SCRCHANGE_SPERM | SCRCHANGE_FROMAPPLY   | SCRCHANGE_TOFULL;
  else         code=SCRCHANGE_SPERM | SCRCHANGE_FROMCONTROL | SCRCHANGE_TOPREVIEW;
  TScrChangeRouter cr;
  cr.hsrcwnd=hsrcwnd; cr.lParam=(LPARAM)this; cr.code=code;
  HWND hdlg=GetParent(hsrcwnd);
  SendMessage(hdlg,SCRM_CHANGE,0,(LPARAM)&cr);
}

void TSpermData::WriteRegistry()
{ LONG res; HKEY skey; DWORD val; DWORD disp;
  res=RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Sperm",0,NULL,
                     REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&skey,&disp);
  if (res!=ERROR_SUCCESS) return;
  // 1. count index
  val=CountIndex;
  RegSetValueEx(skey,"CountIndex",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 2. erase
  val=Erase;
  RegSetValueEx(skey,"Erase",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  // 3. flat out
  val=FlatOut;
  RegSetValueEx(skey,"FlatOut",0,REG_DWORD,(CONST BYTE*)&val,sizeof(val));
  RegCloseKey(skey);
}

void TSpermData::ReadRegistry()
{ LONG res; HKEY skey; DWORD val,valsize,valtype;
  res=RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\Lu\\Sperm",0,KEY_ALL_ACCESS,&skey);
  if (res!=ERROR_SUCCESS) return;
  // 1. count index
  valsize=sizeof(val); res=RegQueryValueEx(skey,"CountIndex",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) CountIndex=val;
  // 2. erase
  valsize=sizeof(val); res=RegQueryValueEx(skey,"Erase",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) Erase=(BOOL) val;
  // 3. flat out
  valsize=sizeof(val); res=RegQueryValueEx(skey,"FlatOut",0,&valtype,(LPBYTE)&val,&valsize);
  if (res==ERROR_SUCCESS) FlatOut=(BOOL) val;
  RegCloseKey(skey);
}


BOOL WINAPI MyScreenSaverConfigureDialog(HWND hwnd, UINT msg,WPARAM,LPARAM lParam)
{ TSpermData *sd=(TSpermData *)GetWindowLong(hwnd,DWL_USER);
  switch (msg) {
    case WM_INITDIALOG:
    { TSpermData *sd=new TSpermData();
		sd->ReadRegistry();
      sd->Mark();
      SetWindowLong(hwnd,DWL_USER,(LONG)sd);
      sd->SetControls(hwnd);
      return TRUE;
    }
    case WM_DESTROY:
    { delete sd;
      SetWindowLong(hwnd,DWL_USER,0);
      return TRUE;
    }
    case WM_COMMAND:
    { HWND hctrlwnd=(HWND) lParam;
      if (hctrlwnd==NULL) return FALSE;
      BOOL ischanged=sd->GetControls(hwnd);
      if (ischanged) sd->NotifyChanges(GetParent(hwnd),FALSE);
      return TRUE;
	 }
	 case (WM_NOTIFY):
    { LPNMHDR nmh=(LPNMHDR) lParam;
      UINT code=nmh->code;
      switch (code) {
        case (PSN_APPLY):
        { sd->Mark();
          sd->WriteRegistry();
          sd->NotifyChanges(GetParent(hwnd),TRUE);
          return TRUE;
        }
        case (PSN_RESET):
        { // Revert back to old values
          BOOL anychanges=sd->Reset();
          if (anychanges) sd->NotifyChanges(hwnd,FALSE);
          return TRUE;
        }
        default: return FALSE;
      }
    }

  }
  return FALSE;
}


int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE,LPSTR,int)
{
  TScrPrefs *p=ScrCreatePrefs();
  p->AllowActiveConfig=TRUE;
  p->AllowMute=FALSE;
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
#ifdef STATICCODE
  strcpy(p->StaticCode,STATICCODE);
#endif
#ifdef DYNAMICCODE
  strcpy(p->DynamicCode,DYNAMICCODE);
#endif
  ScrSetPrefs(p);
  int res=ScrExecute(hInstance,MyScreenSaverProc,MyScreenSaverConfigureDialog);
  return res;
}


